"""Core functions."""
from functools import wraps


def singleton(cls):
    """Ensure singleton instance."""
    instances = {}

    @wraps(cls)
    def instance(*args, **kwargs):
        """Create class instance."""
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return instance



import json
import logging
import os
import random
import re
from pathlib import Path
from typing import Any, List

from cleo import Command
from dotenv import load_dotenv

from config.exceptions import RequestFailedException
from config.spring import ConfigClient

logging.disable(logging.ERROR)


env_path = Path(".") / ".env"
load_dotenv(dotenv_path=env_path)


class CloudFoundryCommand(Command):
    """
    Interact with CloudFoundry via cli.

    cf
    """

    def handle(self):
        pass


class ConfigClientCommand(Command):
    """
    Interact with Spring Cloud Server via cli.

    client
        {app : Application name.}
        {filter? : Config selector.}
        {--a|address=http://localhost:8888 : ConfigServer address.}
        {--b|branch=master : Branch config.}
        {--p|profile=development : Profile config.}
        {--u|url : Base URL format. <option=bold>(default: "<address>/<branch>/<app>-<profile>")</>}
        {--file : Gets remote file from server and saves locally.}
        {--json : Save output as json.}
        {--all : Show all config.}
    """

    EMOJI_ERRORS: List[str] = [
        "\U0001f92f",
        "\U0001f635",
        "\U0001f92e",
        "\U0001f922",
        "\U0001f628",
        "\U0001f62d",
        "\U0001f4a9",
        "\U0001f494",
        "\U0001f4a5",
        "\U0001f525",
    ]
    EMOJI_SUCCESS: List[str] = [
        "\U0001f973",
        "\U0001f929",
        "\U0001f63b",
        "\U0001f496",
        "\U0001f389",
        "\U0001f38a",
    ]
    EMOJI_NOT_FOUND = [
        "\U0001f642",
        "\U0001f60c",
        "\U0001f928",
        "\U0001f643",
        "\U0001f605",
    ]

    def handle(self) -> None:
        filter_options = self.argument("filter") or ""
        host = os.getenv("CONFIGSERVER_ADDRESS", self.option("address"))
        url = os.getenv("CONFIGSERVER_CUSTOM_URL")
        if not url:
            url = f"{host}/{self.option('branch')}/{self.argument('app')}-{self.option('profile')}.json"

        client = ConfigClient(
            address=os.getenv("CONFIGSERVER_ADDRESS", self.option("address")),
            branch=os.getenv("BRANCH", self.option("branch")),
            app_name=os.getenv("APP_NAME", self.argument("app")),
            profile=os.getenv("PROFILE", self.option("profile")),
            url=self.option("url") or url,
            fail_fast=False,
        )

        if self.option("file"):
            self.request_file(client, filter_options)
            raise SystemExit(0)

        content = self.request_config(client, filter_options)
        if self.option("json"):
            self.save_file("output.json", content)
        else:
            self.std_output(filter_options, content)

    def request_config(self, client: ConfigClient, filter_options: str) -> Any:
        self.line("<options=bold>\U000023f3 contacting server...</>")
        try:
            client.get_config()
        except ConnectionError:
            emoji = random.choice(self.EMOJI_ERRORS)
            self.line(f"<options=bold>{emoji} failed to contact server... {emoji}</>")
            raise SystemExit(1)

        self.print_contact_server_ok()
        content = self.get_config(client, filter_options)
        self.has_content(content, filter_options)
        return content

    def request_file(self, client: ConfigClient, filter_options: str) -> None:
        self.line("<options=bold>\U000023f3 contacting server...</>")
        try:
            response = client.get_file(filter_options)
        except RequestFailedException:
            emoji = random.choice(self.EMOJI_ERRORS)
            self.line(f"<options=bold>{emoji} failed to contact server... {emoji}</>")
            raise SystemExit(1)
        with open(f"{filter_options}", "w") as f:
            f.write(response)
        self.line(f"file saved: <info>{filter_options}</info>")

    def get_config(self, client: ConfigClient, filter_options: str) -> Any:
        if self.option("all"):
            content = client.config
        else:
            content = client.get_attribute(f"{filter_options}")
        return content

    def print_contact_server_ok(self) -> None:
        emoji = random.choice(self.EMOJI_SUCCESS)
        self.line(f"<options=bold>{emoji} Ok! {emoji}</>")

    def has_content(self, content, filter_options: str) -> None:
        if len(str(content)) == 0:
            emoji = random.choice(self.EMOJI_NOT_FOUND)
            self.line(
                f"{emoji} no result found for your filter: <comment>'{filter_options}'</comment>"
            )
            raise SystemExit(0)

    def std_output(self, filter_options: str, content: str) -> None:
        if self.option("all"):
            filter_options = "all"
        self.line(
            f"<options=bold>\U0001f4c4 report for filter: <comment>'{filter_options}'</comment>:</>"
        )
        self.line(f"{json.dumps(content, indent=4, sort_keys=True)}")

    def save_file(self, filename: str, content: str) -> None:
        extension = filename[-4:]
        self.line(f"generating <info>{extension}</info> file...")
        with open(f"{filename}", "w") as f:
            json.dump(content, f, indent=4, sort_keys=True)
        self.line(f"file saved: <info>{filename}</info>")


class DecryptCommand(Command):
    """
    Decrypt a input via Spring Cloud Config.

    decrypt
        {data : Data to decrypt.}
        {--a|address=http://localhost:8888 : ConfigServer address.}
        {--p|path=/decrypt : decrypt path.}
    """

    def handle(self):
        client = ConfigClient(
            address=os.getenv("CONFIGSERVER_ADDRESS", self.option("address")),
            fail_fast=False,
        )
        try:
            data = re.match(r"^.?{cipher}?(?P<name>\w.*)", self.argument("data")).group(
                "name"
            )
        except AttributeError:
            data = self.argument("data")
        try:
            resp = client.decrypt(data, path=self.option("path"))
        except Exception:
            self.line("<options=bold>failed to contact server... </>")
            raise SystemExit(1)
        self.line(resp)


class EncryptCommand(Command):
    """
    Encrypt a input via Spring Cloud Config.

    encrypt
        {data : Data to encrypt.}
        {--a|address=http://localhost:8888 : ConfigServer address.}
        {--p|path=/encrypt : encrypt path.}
        {--raw=no : Format output including {cipher}?}
    """

    def handle(self):
        client = ConfigClient(
            address=os.getenv("CONFIGSERVER_ADDRESS", self.option("address")),
            fail_fast=False,
        )
        try:
            resp = client.encrypt(self.argument("data"), path=self.option("path"))
        except Exception:
            self.line("<options=bold>failed to contact server... </>")
            raise SystemExit(1)
        if not self.option("raw") == "yes":
            self.line(f"'{{cipher}}{resp}'")
        else:
            self.line(resp)



from cleo import Application

from config.__version__ import __version__
from config.cli import (
    CloudFoundryCommand,
    ConfigClientCommand,
    DecryptCommand,
    EncryptCommand,
)

application = Application("config-client", f"{__version__}")
application.add(CloudFoundryCommand())
application.add(ConfigClientCommand())
application.add(EncryptCommand())
application.add(DecryptCommand())


if __name__ == "__main__":
    application.run()



import json
import os
from typing import Any

import attr
from glom import glom

from config.cloudfoundry import default_vcap_application, default_vcap_services


@attr.s(slots=True)
class CFenv:
    vcap_service_prefix = attr.ib(
        type=str,
        default=os.getenv("VCAP_SERVICE_PREFIX", "p-config-server"),
        validator=attr.validators.instance_of(str),
    )
    vcap_application = attr.ib(
        type=dict,
        default=json.loads(os.getenv("VCAP_APPLICATION", default_vcap_application)),
        validator=attr.validators.instance_of(dict),
    )
    vcap_services = attr.ib(
        type=dict,
        default=json.loads(os.getenv("VCAP_SERVICES", default_vcap_services)),
        validator=attr.validators.instance_of(dict),
    )

    def __attrs_post_init__(self) -> None:
        if self.vcap_service_prefix not in self.vcap_services.keys():
            vcap_services_copy = self.vcap_services.copy()
            vcap_services_copy[self.vcap_service_prefix] = vcap_services_copy.pop(
                "p-config-server"
            )
            self.vcap_services = vcap_services_copy

    @property
    def space_name(self) -> Any:
        return glom(self.vcap_application, "space_name", default="")

    @property
    def organization_name(self) -> Any:
        return glom(self.vcap_application, "organization_name", default="")

    @property
    def application_name(self) -> Any:
        return glom(self.vcap_application, "application_name", default="")

    @property
    def uris(self) -> Any:
        return glom(self.vcap_application, "uris", default=[])

    def configserver_uri(self, vcap_path: str = ".0.credentials.uri") -> Any:
        path = self._format_vcap_path(vcap_path)
        return glom(self.vcap_services, path, default="")

    def configserver_access_token_uri(
        self, vcap_path: str = ".0.credentials.access_token_uri"
    ) -> Any:
        path = self._format_vcap_path(vcap_path)
        return glom(self.vcap_services, path, default="")

    def configserver_client_id(
        self, vcap_path: str = ".0.credentials.client_id"
    ) -> Any:
        path = self._format_vcap_path(vcap_path)
        return glom(self.vcap_services, path, default="")

    def configserver_client_secret(
        self, vcap_path: str = ".0.credentials.client_secret"
    ) -> Any:
        path = self._format_vcap_path(vcap_path)
        return glom(self.vcap_services, path, default="")

    def _format_vcap_path(self, path: str) -> str:
        return f"{self.vcap_service_prefix}{path}"



import attr
import requests
from requests.auth import HTTPBasicAuth

from config import logger
from config.exceptions import RequestFailedException, RequestTokenException


@attr.s(slots=True)
class OAuth2:
    access_token_uri = attr.ib(type=str, validator=attr.validators.instance_of(str))
    client_id = attr.ib(type=str, validator=attr.validators.instance_of(str))
    client_secret = attr.ib(type=str, validator=attr.validators.instance_of(str))
    grant_type = attr.ib(
        type=str,
        default="client_credentials",
        validator=attr.validators.instance_of(str),
    )
    _token = attr.ib(type=str, default="", validator=attr.validators.instance_of(str))

    @property
    def token(self) -> str:
        return self._token

    @property
    def authorization_header(self) -> dict:
        return {"Authorization": f"Bearer {self._token}"}

    def request_token(self, client_auth: HTTPBasicAuth, data: dict) -> None:
        try:
            response = requests.post(self.access_token_uri, auth=client_auth, data=data)
            if response.ok:
                self._token = response.json().get("access_token")
                logger.info("Access token successfully obtained.")
                logger.debug(f"access_token: {self._token}")
            else:
                raise RequestTokenException(
                    "Failed to retrieve oauth2 access_token. "
                    f"HTTP Response code: {response.status_code}."
                )
        except requests.exceptions.MissingSchema as err:
            logger.error(err)
            raise RequestFailedException("Access token URI it's empty")

    def configure(self) -> None:
        client_auth = HTTPBasicAuth(self.client_id, self.client_secret)
        data = {"grant_type": f"{self.grant_type}"}
        self.request_token(client_auth, data)



import requests


def request(uri: str, **kwargs) -> requests.Response:
    response = requests.get(uri, **kwargs)
    response.raise_for_status()
    return response


def post(uri: str, data: str, **kwargs) -> requests.Response:
    response = requests.post(uri, data=data, **kwargs)
    response.raise_for_status()
    return response



import json

default_vcap_services = json.dumps(
    {
        "p-config-server": [
            {
                "credentials": {
                    "uri": "",
                    "access_token_uri": "",
                    "client_id": "",
                    "client_secret": "",
                }
            }
        ]
    }
)
default_vcap_application = json.dumps(
    {"application_name": "", "space_name": "", "organization_name": "", "uris": []}
)



from aiohttp import web
from glom import glom

from config.spring import ConfigClient


class AioHttpConfig:
    def __init__(
        self, app: web.Application, key: str = "config", client=ConfigClient()
    ) -> None:
        self._validate_app(app)
        self._validate_client(client)
        client.get_config()
        app[str(key)] = _Config(client.config)

    def _validate_app(self, app: web.Application) -> None:
        if not isinstance(app, web.Application):
            raise TypeError("instance must be <aiohttp.web.Application>")

    def _validate_client(self, client) -> None:
        if client.__class__.__name__ not in ("ConfigClient", "CF"):
            raise TypeError("instance must be <ConfigClient> or <CF>")


class _Config(dict):
    def get(self, key, default=None):
        return glom(self, key, default=default)



from typing import Any, Dict, KeysView

import attr

from config.auth import OAuth2
from config.cfenv import CFenv
from config.spring import ConfigClient


@attr.s(slots=True)
class CF:
    cfenv = attr.ib(
        type=CFenv,
        factory=CFenv,
        validator=attr.validators.instance_of(CFenv),
    )
    oauth2 = attr.ib(type=OAuth2, default=None)
    client = attr.ib(type=ConfigClient, default=None)

    def __attrs_post_init__(self) -> None:
        if not self.oauth2:
            self.oauth2 = OAuth2(
                access_token_uri=self.cfenv.configserver_access_token_uri(),
                client_id=self.cfenv.configserver_client_id(),
                client_secret=self.cfenv.configserver_client_secret(),
            )

        if not self.client:
            self.client = ConfigClient(
                address=self.cfenv.configserver_uri(),
                app_name=self.cfenv.application_name,
                profile=self.cfenv.space_name.lower(),
                oauth2=self.oauth2,
            )

    @property
    def vcap_services(self):
        return self.cfenv.vcap_services

    @property
    def vcap_application(self):
        return self.cfenv.vcap_application

    def get_config(self, **kwargs) -> None:
        self.client.get_config(**kwargs)

    @property
    def config(self) -> Dict:
        return self.client.config

    def get_attribute(self, value: str) -> Any:
        return self.client.get_attribute(value)

    def __getitem__(self, key):
        return self.client[key]

    def get_keys(self) -> KeysView:
        return self.client.get_keys()

    def keys(self) -> KeysView:
        return self.client.get_keys()



from typing import Any

import attr
from flask.app import Flask
from flask.config import Config
from glom import glom

from config.spring import ConfigClient


def _validate(instance, attribute, value) -> None:
    client = instance._client
    if client.__class__.__name__ not in ("ConfigClient", "CF"):
        raise TypeError(f"instance must be <ConfigClient> or <CF> (got: {client})")


@attr.s(frozen=True)
class FlaskConfig:
    app = attr.ib(type=Flask, validator=attr.validators.instance_of(Flask), repr=False)
    _client = attr.ib(factory=ConfigClient, validator=_validate)

    def __attrs_post_init__(self) -> None:
        self._client.get_config()
        self.app.config = _Config(
            root_path=self.app.root_path,
            defaults=self.app.config,
            _config=self._client.config,
        )


class _Config(Config):
    def __init__(self, _config, *args, **kwargs):
        super(_Config, self).__init__(*args, **kwargs)
        Config.update(self, _config)

    def get(self, key, default=None) -> Any:
        return glom(self, key, default=default)



from config import __version__

from collections import OrderedDict

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="config-client",
    version=f"{__version__}",
    author="alexandre menezes",
    author_email="alexandre.fmenezes@gmail.com",
    description="config service client for Spring Cloud Config Server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    url="https://github.com/amenezes/config-client",
    packages=setuptools.find_packages(include=["config", "config.*"]),
    python_requires=">=3.6.0",
    project_urls=OrderedDict(
        (
            ("Documentation", "https://config-client.amenezes.net"),
            ("Code", "https://github.com/amenezes/config-client"),
            ("Issue tracker", "https://github.com/amenezes/config-client/issues"),
        )
    ),
    install_requires=["attrs<=20.2.0", "glom<=19.10.0", "requests>=2.22.0"],
    tests_require=[
        "pytest==6.0.2",
        "flake8==3.8.2",
        "flake8-blind-except==0.1.1",
        "flake8-polyfill==1.0.2",
        "pytest-cov==2.10.1",
        "pytest-mock==3.3.1",
        "isort==5.5.3",
        "black==20.8b1",
        "mypy>=0.782",
        "python-dotenv>=0.14.0",
        "aiohttp>=3.5.4",
        "flask>=1.0.0",
    ],
    extras_require={
        "cli": ["cleo>=0.7.6", "python-dotenv>=0.10.3"],
        "docs": ["portray>=1.3.1"],
        "all": ["cleo>=0.7.6", "python-dotenv>=0.10.3", "portray>=1.3.1"],
    },
    setup_requires=["setuptools>=38.6.0"],
    entry_points={"console_scripts": ["config=config.__main__:application.run [cli]"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Java Libraries",
    ],
)



"""Module for retrieve application's config from Spring Cloud Config."""
import os

# from contextlib import suppress
from distutils.util import strtobool
from functools import wraps
from typing import Any, Callable, Dict, KeysView, Optional

import attr
from glom import glom

from config import http, logger
from config.auth import OAuth2
from config.core import singleton
from config.exceptions import RequestFailedException


@attr.s(slots=True)
class ConfigClient:
    """Spring Cloud Config Client."""

    address = attr.ib(
        type=str,
        default=os.getenv("CONFIGSERVER_ADDRESS", "http://localhost:8888"),
        validator=attr.validators.instance_of(str),
    )
    branch = attr.ib(
        type=str,
        default=os.getenv("BRANCH", "master"),
        validator=attr.validators.instance_of(str),
    )
    app_name = attr.ib(
        type=str,
        default=os.getenv("APP_NAME", ""),
        validator=attr.validators.instance_of(str),
    )
    profile = attr.ib(
        type=str,
        default=os.getenv("PROFILE", "development"),
        validator=attr.validators.instance_of(str),
    )
    url = attr.ib(
        type=str,
        default="{address}/{branch}/{app_name}-{profile}.json",
        validator=attr.validators.instance_of(str),
    )
    fail_fast = attr.ib(
        type=bool,
        default=bool(strtobool(str(os.getenv("CONFIG_FAIL_FAST", True)))),
        validator=attr.validators.instance_of(bool),
        converter=bool,
    )
    oauth2 = attr.ib(
        type=Optional[OAuth2],
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(OAuth2)),
    )
    _config = attr.ib(
        type=dict,
        factory=dict,
        init=False,
        validator=attr.validators.instance_of(dict),
        repr=False,
    )

    def __attrs_post_init__(self) -> None:
        """Format ConfigClient URL."""
        self.url = self.url.format(
            address=self.address,
            branch=self.branch,
            app_name=self.app_name,
            profile=self.profile,
        )
        self._ensure_request_json()

    def _ensure_request_json(self) -> None:
        """Ensure that the URI to retrieve the configuration will have the .json extension"""
        if not self.url.endswith(".json"):
            dot_position = self.url.rfind(".")
            if dot_position > 0:
                self.url = self.url.replace(self.url[dot_position:], ".json")
            else:
                self.url = f"{self.url}.json"
            logger.warning(
                "URL suffix adjusted to a supported format. "
                "For more details see: "
                "https://config-client.amenezes.net/docs/1.-overview/#default-values"
            )
        logger.debug(f"Target URL configured: {self.url}")

    def get_config(self, **kwargs: dict) -> None:
        """Request the configuration from the config server."""
        kwargs = self._configure_oauth2(**kwargs)
        try:
            response = http.request(self.url, **kwargs)
        except Exception as ex:
            logger.error(f"Failed to request: {self.url}")
            logger.error(ex)
            if self.fail_fast:
                logger.info("fail_fast enabled. Terminating process.")
                raise SystemExit(1)
            raise ConnectionError("fail_fast disabled.")
        self._config = response.json()

    def _configure_oauth2(self, **kwargs):
        if self.oauth2:
            self.oauth2.configure()
            try:
                kwargs["headers"].update(self.oauth2.authorization_header)
            except KeyError:
                kwargs.update(dict(headers=self.oauth2.authorization_header))
        return kwargs

    def get_file(self, filename: str, **kwargs: dict) -> str:
        """Request a file from the config server."""
        uri = f"{self.address}/{self.app_name}/{self.profile}/{self.branch}/{filename}"
        try:
            response = http.request(uri, **kwargs)
        except Exception:
            raise RequestFailedException(f"Failed to request URI: {uri}")
        return response.text

    def encrypt(
        self,
        value: str,
        path: str = "/encrypt",
        headers: dict = {"Content-Type": "text/plain"},
        **kwargs: dict,
    ) -> str:
        """Request a encryption from a value to the config server."""
        try:
            response = http.post(
                uri=f"{self.address}{path}", data=value, headers=headers, **kwargs
            )
        except Exception:
            raise RequestFailedException(f"Failed to request URI: {self.address}{path}")
        return response.text

    def decrypt(
        self,
        value: str,
        path: str = "/decrypt",
        headers: dict = {"Content-Type": "text/plain"},
        **kwargs: dict,
    ) -> str:
        """Request a decryption from a value to the config server.."""
        try:
            response = http.post(
                uri=f"{self.address}{path}", data=value, headers=headers, **kwargs
            )
        except Exception:
            raise RequestFailedException(f"Failed to request URI: {self.address}{path}")
        return response.text

    @property
    def config(self) -> Dict:
        """Getter from configurations retrieved from ConfigClient."""
        return self._config

    def __getitem__(self, key: str) -> Any:
        return glom(self._config, key, default="")

    def keys(self):
        return self._config.keys()

    def get_attribute(self, value: str) -> Any:
        """
        Get attribute from configuration.

        :value value: -- The filter to query on dict.
        :return: The value matches or ''
        """
        logger.warning(
            "get_attribute it's deprecated, access attribute directly now like: spring_config['my.config']."
        )
        return glom(self._config, value, default="")

    def get_keys(self) -> KeysView:
        """List all keys from configuration retrieved."""
        logger.warning("get_keys it's deprecated, please use keys() method.")
        return self._config.keys()


@singleton
def create_config_client(**kwargs) -> ConfigClient:
    """
    Create ConfigClient singleton instance.

    :param address:
    :param app_name:
    :param branch:
    :param fail_fast:
    :param profile:
    :param url:
    :return: ConfigClient instance.
    """
    obj = ConfigClient(**kwargs)
    obj.get_config()
    return obj


def config_client(**kwargs) -> Callable[[Dict[str, str]], ConfigClient]:
    """ConfigClient decorator.

    Usage:

    @config_client(app_name='test')
    def get_config(config):
        db_user = config.get_attribute('database.user')

    :raises: ConnectionError: If fail_fast enabled.
    :return: ConfigClient instance.
    """
    logger.debug("kwargs: %r", kwargs)

    cls_attributes = attr.fields_dict(ConfigClient).keys()
    instance_params = {}
    get_config_params = {}
    for key, value in kwargs.items():
        if key in cls_attributes:
            instance_params.update({key: value})
        else:
            get_config_params.update({key: value})

    def wrap_function(function):
        logger.debug("caller: %s", function.__name__)

        @wraps(function)
        def enable_config():
            obj = ConfigClient(**instance_params)
            obj.get_config(**get_config_params)
            return function(obj)

        return enable_config

    return wrap_function
